static char query_contents_c_sourceCodeID[]="$Id: @(#)03  1.4.1.3 src/pdweb/webseald/junctions/remote/query_contents/query_contents.c, pdweb.ws.junctions.remote, amweb611 06/09/26 15:01:47 @(#) $";

/*
 *	query_contents.c
 *
 *	An implementation of the CGI program for determining the contents of
 *	the web space of a third party web server.  Returns multi-line
 *	plain text output in the following format:
 *
 *	content-type: text/plain;charset=utf8
 *
 *	return-code
 *	files/directories as applicable, one per line.
 *
 *	This implementation is particular to the WIN/32 platform, although
 *	it could be easily adapted to other platforms not supporting a UNIX
 *	shell environment.
 *
 *	No support currently exists for directory mapping.  This could be added by
 *	looking up the registry entries for a particular web server, and incorporating
 *	this into the directory_list function.
 *
 *  Command Line Format:
 *
 *		query_contents dirlist=dir
 *	or
 *		export QUERY_STRING="dirlist=<dirpath>"
 *		query_contents
 *
 *	History:
 *
 *	09-13-96	created
 *	06/02/2006	handled a uri-encoded UTF-8 query string.
 */

/* Notes:   2001/07/31
 * As a cgi, it's reasonably important that we don't intruduce unwanted security holes
 * So, we don't allow '.'s in paths, we don't parse %xy or Unicode definitions
 * and it won't compile on Unix until you set the webserver document root path
 * Don't GUESS, make SURE you get that right.
 * Policy Director doesn't need this to do cute things, we only look through the
 * web space from root to leaves so we can check that objects exist in the web space.
 */

/* Notes:   2006/06/02  LG
 * This version of query_contents.c is a Windows-only source file.
 * A query_contents executable built from this file is shipped and
 * used for Windows only, so the Unix-specific code has been removed.
 * To see the Unix version of the code look at the earlier versions.
 *
 * When query_contents is built in Msdev for Debug then the executable
 * cannot be used as a cgi program.  Extensive debug printfs are
 * enabled in such a build making the program suitable for execution
 * from a command line only.  Unable to find a way to run Msdev or
 * Windbg in a Chinese locale but using a C locale UI, I used these printfs
 * to debug the code in the Chinese locale.
 *
 * The uri-encoded UTF-8 query string arises only in a MBCS locale where
 * non-ASCII file and directory names are used.  This code should make no
 * difference for 99.8% of our customers.  With the bounds checking on
 * buffers there should be no security problem with "%xx" encoded characters.
 *
 * Instead of rejecting any path containing a '.' the new code rejects any
 * path containing a "..".  This should prevent a malicous user of
 * query_contents from examing files outside of the docroot tree but still
 * allow a directory name of "a.b".
 *
 * The uri-encoding of the query string is now optional and under the control
 * of the hidden config option "query-contents-uriencoded" (which defaults to
 * "no" for backward compatability).  A cookie, x-query-contents-uriencoded,
 * is now sent to query_contents so it can know whether or not to uri decode
 * the query string it receives.  If no cookie is received then no uri
 * decoding is done.
 *
 */


#define UNICODE
#include <windows.h>
#include <stdio.h>
#include <string.h>


/* error codes */
#define EXIT_OK			"100\n"
#define EXIT_USAGE		"101\n"
#define EXIT_NO_OBJ		"102\n"
#define EXIT_NO_DIR		"103\n"
#define EXIT_CANT_CONVERT_TO_UNICODE "104\n"

/*
 * The environment variable that is set if WebSEAL sends
 * the x-query-contents-uriencoded cookie.
 */
#define URIDECODE_COOKIE    "HTTP_X_QUERY_CONTENTS_URIENCODED"


/* local prototypes */
static void parse_arguments( char * src, wchar_t * lhs, wchar_t * rhs );
static void directory_list( wchar_t * object );
static void object_check( char * object );
static wchar_t *get_docroot( void );
static wchar_t *full_object_path( wchar_t * object );
static void fwdslash_to_backslash( wchar_t * string );
static char * uri_decode( const char * string );
#ifdef _DEBUG
    static char * uri_encode( const char * string );
#endif


/*
 *	parse_arguments
 *		Build up the command and name space object strings.  This function
 *  	will exit if it cannot find these strings properly.
 *
 *	PARAMETERS
 *		src		The argument as passed to the program from the command line or
 *				QUERY_STRING environment variable.
 *		lhs		The lhs result of parsing
 *		rhs		The rhs result of parsing
 *
 *	RETURNS
 *		nothing
 */
static void
parse_arguments( char *src, wchar_t *lhs, wchar_t *rhs )
{
    int       nWideChars;
    int	      status;
    wchar_t * lpszUnicodeSource;
    wchar_t * lpwEq;
    char    * uridecode_cookie_value = getenv(URIDECODE_COOKIE);

#ifdef _DEBUG
    wchar_t * pwc;
    char    * pc;
#endif

    if ( src == (char *)NULL ) {
	printf( EXIT_NO_DIR );
	exit(0);
    }

    if (uridecode_cookie_value && !strcmp(uridecode_cookie_value, "yes")) {
        /*
         *  The query string is uri-encoded so uri-decode it first.
         *  Don't worry about memory leaks - we won't live long
         *  enough for that.
         */
    #ifdef _DEBUG
        printf ("source argument: \"%s\"\n", src);
    #endif
        src = uri_decode(src);
    #ifdef _DEBUG
        printf ("uri-decoded arg: ");
        for (pc = src; *pc; pc++) {
            printf (" %02x", *pc);
        }
        printf ("\n");
    #endif
    }

    /*
     *  Now that we've uri-decoded the argument into UTF-8 characters
     *  create a Unicode (UTF-16) version.  First get the length
     *  of the buffer we'll need, then create it and do the translation
     *  into it.
     */
    nWideChars = MultiByteToWideChar(CP_UTF8, 0, src, -1, NULL, 0);
    if (nWideChars == 0) {
	printf( EXIT_CANT_CONVERT_TO_UNICODE );
	exit(0);
    }
    lpszUnicodeSource = malloc(sizeof(wchar_t)*(nWideChars + 1));
    status = MultiByteToWideChar(CP_UTF8, 0, src,
    				 -1, lpszUnicodeSource, nWideChars);
    if (status == 0) {
	printf( EXIT_CANT_CONVERT_TO_UNICODE);
	exit(0);
    }

#ifdef _DEBUG
    printf("MultiByte Unicode string: ");
    for (pwc=lpszUnicodeSource; *pwc; pwc++) {
        printf (" %04x", *pwc);
    }
    printf ("\n");
#endif

    /*
     *  Now find the '=' and split the argument string into two pieces,
     *  complaining if we find no '='.
     */
    lpwEq = wcschr(lpszUnicodeSource, L'=');
    if (lpwEq == 0) {
	printf( EXIT_NO_DIR );
	exit(0);
    }
    *lpwEq++ = 0;
    wcsncpy(lhs, lpszUnicodeSource, MAX_PATH);
    wcsncpy(rhs, lpwEq, MAX_PATH);

    /*
     *  Now convert all forward slashes to back slashes for the Windows
     *  directory lookup.
     */
    // fwdslash_to_backslash( rhs );

    free(lpszUnicodeSource);
}


/*
 *	directory_list
 *		Lists objects in the directory specified in uri-encoded UTF-8.
 *		The directory is relative (or absolute to the "document root").
 *
 *	PARAMETERS
 *		object	A path relative to the document root.  This is converted
 *			into an absolute path for use within this function.
 *
 *	RETURNS
 *		none
 */
static void
directory_list( wchar_t *object )
{
    wchar_t		* fop = full_object_path( object );
    WIN32_FIND_DATA	  win32fData;
    HANDLE		  win32fHandle;
    char	        * lpUTF8FileName = NULL;
    int			  nUTF8Bytes;
    int			  status;

    wcsncat( fop, L"\\*.*", (size_t)(MAX_PATH - wcslen(fop)));
    win32fHandle = FindFirstFile( fop, &win32fData );
    if ( win32fHandle == INVALID_HANDLE_VALUE ) {
	printf( EXIT_NO_DIR );
	exit (0);
    } else {
	printf( EXIT_OK );
    }

    while ( 1 ) {
	/*
	 *  Ignoring "." and "..", print each file/directory name found
	 *  in the specified directory.  Each name is translated from
	 *  Unicode to UTF-8, uri-endcoded, then printed.
	 *  A double-slash is appended if it's a directory.
	 */
	if ( wcscmp( win32fData.cFileName, L"." ) &&
	     wcscmp( win32fData.cFileName, L".." ) ) {
#ifdef _DEBUG
            wchar_t * pwc = NULL;
            char * pc = NULL;
            printf ("Unicode win32fData.cFileName: ");
            for (pwc = win32fData.cFileName; *pwc; pwc++) {
                printf (" %04x", *pwc);
            }
            printf ("\n");
#endif
	    nUTF8Bytes = WideCharToMultiByte(CP_UTF8, 0, win32fData.cFileName, -1,
	    				     lpUTF8FileName, 0, NULL, NULL);
#ifdef _DEBUG
	    printf ("nUTF8Bytes = %d\n", nUTF8Bytes);
#endif
	    lpUTF8FileName = malloc (nUTF8Bytes);
	    status = WideCharToMultiByte(CP_UTF8, 0, win32fData.cFileName, -1,
	    				 lpUTF8FileName, nUTF8Bytes, NULL, NULL);
#ifdef _DEBUG
	    printf ("status returned by WideCharToMultiByte(): %d\n", status);
            printf ("MultiByte string: ");
            for (pc = lpUTF8FileName; *pc; pc++) {
                printf (" %02x", (unsigned char *)*pc);
            }
            printf ("\n");

	    uri_encode(lpUTF8FileName);
#endif
            printf ("%s", lpUTF8FileName);
            free(lpUTF8FileName);

	    if ( win32fData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) {
		    /* directories are signified with a trailing double slash */
		    printf( "//" );
	    }
	    printf( "\n" );
	}
	if ( !FindNextFile( win32fHandle, &win32fData ) ) {
	    break;
	}
    }

    FindClose( win32fHandle );
}


/*
 *	get_docroot
 *		Returns a pointer to a string that is the document root for the
 *		web space.  We obtain this from a plain text configuration file on
 *  	        the web server, in the same directory as the query_contents script.
 *		We assume that this file will be in the Windows directory when
 *		this program runs.  The name of this file is query_contents.cfg and
 *		the format is as follows:
 *
 *			[server]
 *			docroot=
 *
 *		It may be desirable to obtain this directly from the registry entry for
 *		the web server in question.  Of course, this would then make the code
 *		web server specific.
 *
 *	PARAMETERS
 *		none
 *
 *	RETURNS
 *		The document root obtained from the configuration file.
 */
static wchar_t *
get_docroot( void )
{
    static wchar_t docroot[MAX_PATH+1];

    GetPrivateProfileString( L"server",
			     L"docroot",
			     L"",
			     docroot,
			     MAX_PATH,
			     L"query_contents.cfg" );
#ifdef _DEBUG
    printf("docroot=[%S]\n",docroot[0] ? docroot: L"docroot is NULL");
#endif
    if ( docroot[0] == '\0' ) {
	printf( EXIT_USAGE );
	exit(0);
    }
	return docroot;
}


/*
 *	full_object_path
 *		Takes a pointer which is the object string and returns a pointer to
 *		the full path of the object, in the WIN/32 filespace.  This function
 *		takes into consideration all of the slash issues along the way.
 *
 *	PARAMETERS
 *		object		Name of the object string, relative to the document root.
 *
 *	RETURNS
 *		The full path of the object on the local file system.
 */
static wchar_t *
full_object_path( wchar_t *object )
{
    static wchar_t full_obj_path[MAX_PATH+1];
    int	    nCharsLeft;

    wcsncpy( full_obj_path, get_docroot(), MAX_PATH );
    nCharsLeft = MAX_PATH - wcslen(full_obj_path);

    if ( full_obj_path[wcslen(full_obj_path)-1] == L'\\' ) {
	if ( object[0] == L'\\' ) {
	    wcsncat( full_obj_path, &object[1], nCharsLeft );
	} else {
	    wcsncat( full_obj_path, object, nCharsLeft );
	}
    } else {
	if ( object[0] == L'\\' ) {
	    wcsncat( full_obj_path, object, nCharsLeft );
	} else {
	    wcsncat( full_obj_path, L"\\", nCharsLeft );
	    wcsncat( full_obj_path, object, nCharsLeft-1 );
	}
    }
    return full_obj_path;
}


/*
 *	fwdslash_to_backslash
 *		Converts from UNIX-style to MS-DOS style directory separators.
 *
 *	PARAMETERS
 *		wchar_t * string	Null-terminated string to convert.
 *					Note that the conversion is done inline.
 *
 */
static void
fwdslash_to_backslash( wchar_t * string )
{
    wchar_t * c;
    for (c = string; *c; c++) {
	if ( *c == L'/' ) {
	    *c = L'\\';
	}
    }
}


/*
 *  Stores the hex digits representing the two nibbles of the
 *  character v into the buffer pointed to by *pOut, advancing
 *  *pOut by 2.
 */
#ifdef _DEBUG
static void hex_encode(char ** pOut, char v)
{
    char lowNibble, highNibble, c;
    printf("  v: %02x\n", v);
    lowNibble = v & 0xf;
    highNibble = (v >> 4) & 0xf;
    printf("  lowNibble = %01x\n", lowNibble);
    printf("  highNibble = %01x\n", highNibble);
    c = highNibble;
    c += c < 10 ? '0' : 'a' - 10;
    printf ("c += c < 10 ? '0' : 'a' - 10: %01x - %c\n", c, c);
    *(*pOut)++ = c;
    c = lowNibble;
    c += c < 10 ? '0' : 'a' - 10;
    printf ("c += c < 10 ? '0' : 'a' - 10: %01x - %c\n", c, c);
    *(*pOut)++ = c;
};


/*
 *  Returns a null-terminated string that is the uri-encoding
 *  of the input argument, which is also a null-terminated string.
 */
static char * uri_encode(const char * v)
{
    int		lenV = strlen(v);
    char * r = malloc(3*lenV+1);
    char * rp = r;
    int i = 0;
    while (i < lenV)
    {
	char c = v[i++];
	if (isalpha(c) || isdigit(c)) {
	    *rp++ = c;
	} else {
	    switch (c)
	    {
	    case ' ':
		*rp++ = '+';
		break;
	    case '-':
	    case ',':
	    case '.':
	    case '@':
	    case '_':
		*rp++ = c;
		break;
	    default:
	    	*rp++ = '%';
		hex_encode(&rp, c);
	    }
	}
    }
    *rp = 0;
    printf ("uri_encode() returning \"%s\"\n", r);
    return r;
};
#endif

/*
 *  Returns the binary value 0 - 15 for which the argument
 *  is the ASCII digit.
 */
static int hex_decode(char v)
{
    v = tolower(v);
    if (v < 'a')
    {
	v -= '0';
	return v < 10 ? v : -1;
    } else {
	v -= 'a';
	return v < 6 ? 10 + v : -1;
    }
};

/*
 *  Returns a null-terminated string that is the uri-decoding
 *  of the input string, which is a uri-encoded null-terminated
 *  string.
 */
static char * uri_decode(const char * pszEncodedString)
{
    char * r = malloc(strlen(pszEncodedString)+1);
    char * pr = r;
    char * psz = (char *)pszEncodedString;
    char   l;

    while (*psz)
    {
	char c = *psz++;
	switch (c)
	{
	case '%':
	    if (!*psz || !*(psz+1) ) {
		return "";
	    }
	    l = *psz++;
	    *pr++ = hex_decode(l) << 4 | hex_decode(*psz++);
	    break;

	case '+':
	    *pr++ = ' ';
	    break;

	default:
	    *pr++ = c;
	}
    }
    *pr = 0;
    return r;
};

/*
 *	main
 *		Entry point to the program.
 *
 */
int
main( int argc, char *argv[] )
{
    static wchar_t command[MAX_PATH+1], nsobject[MAX_PATH+1];

    /* server needs this info for CGI return */
    printf( "content-type: text/plain;charset=utf-8\n\n" );

    /* Ensure that 'command' and 'nsobject' are properly terminated
     * if a wcsncpy() has to copy exactly MAX_PATH wide characters
     * into either one of them.
     */
    command[MAX_PATH] = 0;
    nsobject[MAX_PATH] = 0;

    if ( argc == 2 ) {
    	parse_arguments( argv[1], command, nsobject );
    } else {
	parse_arguments( getenv("QUERY_STRING"), command, nsobject );
    }

    /* branch on command (only one choice) */
    if ( wcscmp( command, L"dirlist" ) == 0 ) {
	wcsncat( nsobject, L"\\", (MAX_PATH - wcslen(nsobject)) );
	/*
	 *  We check for '..' in the input here and simply say NO!
	 *  PD never uses this.
	 */
	if ( wcsstr(nsobject, L"..") != NULL) {
	  printf(EXIT_USAGE);
	  exit(1);
	}
    } else {
	  printf(EXIT_USAGE);
	  exit(0);
    }

    directory_list( nsobject );

    return 0 ;
}

